/*
* FTCS422 Lab #6
* Authors Tom Lynch, Alwyn Nixon-Lloyd
*/

#include <eyebot.h>
#include <math.h>
#include "DCVCtrl.h"

#define NIN 	4
#define NHID	3
#define NOUT	2
#define MAX_DRIVE	160
#define OBSTACLE_THRESHOLD 150

int front_value, right_value, left_value;

PSDHandle leftPSD, rightPSD, frontPSD;
MotorHandle leftMotor, rightMotor;



float N_in[NIN];
float N_hid[NHID];
float N_out[NOUT];

float w_in[NIN][NHID];
float w_out[NHID][NOUT];

float sigmoid(float x) {
	return (1/(1 + exp(-x)));
}

//Maps 0..1 outputs to left and right motor drive values
int interpretNN(float value) {
	int drive;
	drive = (value - 0.5)*MAX_DRIVE;

	//clip motor drive value
	drive = (drive > 100)?100:drive;
	drive = (drive < -100)?(-100):drive;

	return drive;
}


/* Feeds given input conditions into NN and returns output in N_out */
void feedforward (float N_in[NIN], float N_hid[NHID], float N_out[NOUT]) {
	int i, j;
	N_in[NIN- 1] = 1.0; // set bias input neuron

	for (i= 0; i< NHID- 1; i++) // calculate activation of hidden neurons
	{
		N_hid[ i] = 0.0;
		for (j= 0; j < NIN; j++) N_hid[ i] += N_in[j] * w_in[j][i];
		N_hid[i] = sigmoid(N_hid[i]);
	}

	N_hid[NHID- 1] = 1.0; // set bias hidden neuron

	for (i= 0; i< NOUT; i++) // calculate output neurons
	{
		N_out[i] = 0.0;
		for (j= 0; j< NHID; j++) N_out[i] += N_hid[j] * w_out[j][i];
		N_out[i] = sigmoid(N_out[i]);
	}
}

int main() {
	int keyPressed = 0;
	int l,r,i;

	//Set up PSDs and motors
	leftPSD = PSDInit(PSD_LEFT);
	rightPSD = PSDInit(PSD_RIGHT);
	frontPSD = PSDInit(PSD_FRONT);

	leftMotor = MOTORInit(LEFTMOTOR);
	rightMotor = MOTORInit(RIGHTMOTOR);
	MOTORDrive(leftMotor,0);
	MOTORDrive(rightMotor,0);

	PSDStart(leftPSD,1);
	PSDStart(rightPSD,1);
	PSDStart(frontPSD,1);



	//Initialise weights
	//Hidden layer input weights
	w_in[0][0] = -7.152205;
	w_in[0][1] = -0.114524;
	w_in[0][2] = 0.783099;
	w_in[1][0] = 8.661428;
	w_in[1][1] = -8.596476;
	w_in[1][2] = 0.197551;
	w_in[2][0] = -15.320573;
	w_in[2][1] = -10.065578;
	w_in[2][2] = 0.277775;
	w_in[3][0] = 10.702194;
	w_in[3][1] = 13.659206;
	w_in[3][2] = 0.628871;
	//Hidden layer output weights
	w_out[0][0] = -1.096293;
	w_out[0][1] = 9.619850;
	w_out[1][0] = 9.704051;
	w_out[1][1] = 0.740476;
	w_out[2][0] = -3.797630;
	w_out[2][1] = -5.143959;








	//Main loop
	while (keyPressed != KEY4) {
		//Get PSD data and convert to binary obstacle indicators

		left_value = PSDGet(leftPSD);
		right_value = PSDGet(rightPSD);
		front_value = PSDGet(frontPSD);

		LCDClear();
		LCDPrintf(" go cosima go! \n");



		N_in[0] = ( 1.0 - (float)left_value/999.0 );
		N_in[1] = (1.0 - (float)right_value/999.0 );
		N_in[2] = (1.0 - (float)front_value/999.0 );

		for(i=0;i<3;i++){

			if(N_in[i] < 0.50){
				N_in[i] = 0.0;
			}

			else if(N_in[i] > 0.85){
				N_in[i] = 1.0;
			}
			else {
				N_in[i] = N_in[i]*0.78;
			}

		}


		//Obtain NN output
		feedforward(N_in, N_hid, N_out);

		//Interpret as drive values
		l = interpretNN(N_out[0]); //left motor, *1.1 to account for defficencys
		r = interpretNN(N_out[1]);

		//Drive motors
		MOTORDrive(leftMotor, l);
		MOTORDrive(rightMotor, r);

		//Output current output values to LCD
		LCDPrintf("L:%1.2f R:%1.2f F:%1.2f\n",N_in[0], N_in[1], N_in[2]);
		LCDPrintf("Left:  %d\n",l);
		LCDPrintf("Right: %d\n",r);

		keyPressed = KEYRead();
	}

	PSDRelease();

	MOTORRelease(leftMotor);
	MOTORRelease(rightMotor);


	return 1;

}
