/*
* FTCS422 Lab #6
* Authors Tom Lynch, Alwyn Nixon-Lloyd
*/

#include <stdlib.h>
#include <math.h>

#define NIN 	4
#define NHID	3
#define NOUT	2
#define ERROR_MAX	0.0001

//PSDHandle leftPSD, rightPSD, frontPSD;
//MotorHandle leftMotor, rightMotor;

float correctRight = 0;
float correctLeft = 0;

float max_error = 1.0;

int number_of_errors = 0;

float w_in[NIN][NHID];
float w_out[NHID][NOUT];

float sigmoid(float x) {
	return (1/(1 + exp(-x)));
}


float error_Total(float latest_Error){

	if(latest_Error > max_error){

		max_error = latest_Error;
		number_of_errors = 0;
		return max_error;
	}

	if(number_of_errors > 100){
		max_error = latest_Error;
		number_of_errors = 0;
		return max_error;

	}

	number_of_errors++;
	return max_error;

}




//Maps binary input combination to correct wheel output directions
getDriveValues(int left, int right, int front) {
	int input = (left << 2) + (right << 1) + front;

	switch (input) {
		case 0: {
			correctLeft = 1;
			correctRight = 1;
			break;
		}
		case 1: {
			correctLeft = 1;
			correctRight = 0;
			break;
		}
		case 2: {
			correctLeft = 1;
			correctRight = 1;
			break;
		}
		case 3: {
			correctLeft = 0;
			correctRight = 1;
			break;
		}
		case 4: {
			correctLeft = 1;
			correctRight = 1;
			break;
		}
		case 5: {
			correctLeft = 1;
			correctRight = 0;
			break;
		}
		case 6: {
			correctLeft = 1;
			correctRight = 1;
			break;
		}
		case 7: {
			correctLeft = 0;
			correctRight = 0;
			break;
		}
		default: break;
	}
}


/* Feeds given input conditions into NN and returns output in N_out */
void feedforward (float N_in[NIN], float N_hid[NHID], float N_out[NOUT]) {
	int i, j;
	N_in[NIN- 1] = 1.0; // set bias input neuron

	for (i= 0; i< NHID- 1; i++) // calculate activation of hidden neurons
	{
		N_hid[ i] = 0.0;
		for (j= 0; j < NIN; j++) N_hid[ i] += N_in[j] * w_in[j][i]; N_hid[i] = sigmoid(N_hid[i]);
	}

	N_hid[NHID- 1] = 1.0; // set bias hidden neuron

	for (i= 0; i< NOUT; i++) // calculate output neurons
	{
		N_out[i] = 0.0;
		for (j= 0; j< NHID; j++) N_out[i] += N_hid[j] * w_out[j][i]; N_out[i] = sigmoid(N_out[i]);
	}
}

float backprop (float train_in[NIN], float train_out[NOUT]) /* returns current square error value */
{
	int i, j;
	float err_total;
	float N_out[NOUT], err_out[NOUT];
	float diff_out[NOUT];
	float N_hid[NHID], err_hid[NHID], diff_hid[NHID]; // run network, calculate difference to desired output

	feedforward(train_in, N_hid, N_out);
	err_total = 0.0;
	for (i= 0; i< NOUT; i++) {
		err_out[i] = train_out[i]- N_out[i];
		diff_out[i]= err_out[i] * (1.0- N_out[i]) * N_out[i];
		err_total += err_out[i]* err_out[i];
	}

	// update w_out and calculate hidden difference values
	for (i= 0; i< NHID; i++) {
		err_hid[i] = 0.0;
		for (j= 0; j< NOUT; j++) {
			err_hid[i] += err_out[j] * w_out[i][j];
			w_out[i][j] += diff_out[j] * N_hid[i];
		}

		diff_hid[i] = err_hid[i] * (1.0- N_hid[i]) * N_hid[i];
	} // update w_in

	for (i= 0; i< NIN; i++)
		for (j= 0; j< NHID; j++)
			w_in[i][j] += diff_hid[j] * train_in[i];

	return err_total;
}



int main() {
	//randomise weights
	int i,j, l, r, s,randleft,randright,randfront;

	float N_in[NIN];
	float N_hid[NHID];
	float N_out[NOUT];

	float rand_in[4];
	float correct_out[2];

	float running_error;

	float long_term_error;


	rand_in[3] = 1;

	for (i=0; i<NIN; i++)
		for (j=0; j<NHID; j++)
			w_in[i][j] = rand()/((float)RAND_MAX);

	for (i=0; i<NHID; i++)
		for (j=0; j<NOUT; j++)
			w_out[i][j] = rand()/((float)RAND_MAX);

	//generate random input and compare resulting output with correct output
	// - iterate until threshold error level is reached
	do {
		randleft =  (rand() %2);
		randright = (rand() %2);
		randfront = (rand() %2);

		rand_in[0] = (float) (randleft);
		rand_in[1] = (float) (randright);
		rand_in[2] = (float) (randfront);

		getDriveValues(randleft, randright, randfront);

		correct_out[0] = correctLeft;
		correct_out[1] = correctRight;

		printf("%f %f\n",correct_out[0],correct_out[1]);

		running_error = backprop(rand_in, correct_out);
		printf("Current error: %f\n", running_error);

		long_term_error = error_Total(running_error);
		printf("long term error %F \n", long_term_error);

	}
	while (long_term_error > ERROR_MAX);

	//output weights initialisation for EyeBot program to console
	printf("//Initialise weights\n");
	printf("//Hidden layer input weights\n");
	for (i = 0; i<NIN; i++)
		for (j = 0; j <NHID;j++)
			printf("w_in[%d][%d] = %f;\n", i, j, w_in[i][j]);
	printf("//Hidden layer output weights\n");
	for (i = 0; i <NHID; i++)
		for (j = 0; j < NOUT; j++)
			printf("w_out[%d][%d] = %f;\n", i, j, w_out[i][j]);

	printf("\n\n");

	N_in[0]=0.0f;
	N_in[1]=0.0f;
	N_in[2]=0.0f;
	N_in[3]=1.0f;
	feedforward(N_in,N_hid,N_out);
	printf("%f %f\n",N_out[0],N_out[1]);


	N_in[2]=1.0f;
	feedforward(N_in,N_hid,N_out);
	printf("%f %f\n",N_out[0],N_out[1]);


	for (l=0;l<2;l++)
		for (r=0;r<2;r++)
			for (s=0;s<2;s++) {
				N_in[0] = l;
				N_in[1] = r;
				N_in[2] = s;

				feedforward(N_in,N_hid,N_out);

				printf("%d %d %d %1.1f %1.1f\n", (int)N_in[0],(int)N_in[1],(int)N_in[2],N_out[0],N_out[1]);
			}







	return 1;

}
